@@ -6,11 +6,16 @@
 # Copyright (c) 2012 the python-tdbus authors. See the file "AUTHORS" for a
 # complete list.
 
-from tdbus._tdbus import DBUS_BUS_SESSION, DBUS_BUS_SYSTEM
+from tdbus._tdbus import DBUS_BUS_SESSION, DBUS_BUS_SYSTEM, \
+    DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER, DBUS_REQUEST_NAME_REPLY_IN_QUEUE, DBUS_REQUEST_NAME_REPLY_EXISTS, \
+    DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER, \
+    DBUS_INTERFACE_INTROSPECTABLE, DBUS_INTERFACE_PROPERTIES, DBUS_INTERFACE_PEER, \
+    DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS, DBUS_SERVICE_DBUS
 from tdbus.connection import DBusConnection, DBusError
 from tdbus.handler import DBusHandler, method, signal_handler
 from tdbus.select import SimpleDBusConnection
 
+
 try:
     from tdbus.gevent import GEventDBusConnection
 except ImportError:
